# stdlib
from enum import Enum
from typing import Any
from typing import TypeVar

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...store.linked_obj import LinkedObject
from ...types.errors import SyftException
from ...types.result import Err
from ...types.result import as_result
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..context import AuthedServiceContext
from ..output.output_service import ExecutionOutput
from ..policy.policy import InputPolicyValidEnum
from ..policy.policy import OutputPolicy
from ..request.request import Request
from ..request.request import SubmitRequest
from ..request.request import SyncedUserCodeStatusChange
from ..request.request import UserCodeStatusChange
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .user_code import SubmitUserCode
from .user_code import UserCode
from .user_code import UserCodeStatus
from .user_code import UserCodeUpdate
from .user_code import get_code_hash
from .user_code import load_approved_policy_code
from .user_code_stash import UserCodeStash


class HasCodePermissionEnum(str, Enum):
    ACCEPTED = "Has permission"
    DENIED = "Permission denied"


class IsExecutionAllowedEnum(str, Enum):
    ALLOWED = "Execution allowed"
    NO_PERMISSION = "Execution denied: You do not have permission to execute code"
    NOT_APPROVED = "Execution denied: Your code is waiting for approval"
    OUTPUT_POLICY_NONE = "Execution denied: Output policy is not set"
    INVALID_OUTPUT_POLICY = "Execution denied: Output policy is not valid"
    OUTPUT_POLICY_NOT_APPROVED = "Execution denied: Output policy not approved"


@serializable(canonical_name="UserCodeService", version=1)
class UserCodeService(AbstractService):
    stash: UserCodeStash

    def __init__(self, store: DBManager) -> None:
        self.stash = UserCodeStash(store=store)

    @service_method(
        path="code.submit",
        name="submit",
        roles=GUEST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def submit(
        self, context: AuthedServiceContext, code: SubmitUserCode
    ) -> SyftSuccess:
        """Add User Code"""
        user_code = self._submit(context, code, exists_ok=False).unwrap()
        return SyftSuccess(
            message="User Code Submitted", require_api_update=True, value=user_code
        )

    @as_result(SyftException)
    def _submit(
        self,
        context: AuthedServiceContext,
        submit_code: SubmitUserCode,
        exists_ok: bool = False,
    ) -> UserCode:
        """
        Submit a UserCode.

        If exists_ok is True, the function will return the existing code if it exists.

        Args:
            context (AuthedServiceContext): context
            submit_code (SubmitUserCode): UserCode to submit
            exists_ok (bool, optional): If True, return the existing code if it exists.
                If false, existing codes returns Err. Defaults to False.

        Returns:
            Result[UserCode, str]: New UserCode or error
        """
        try:
            existing_code = self.stash.get_by_code_hash(
                context.credentials,
                code_hash=get_code_hash(submit_code.code, context.credentials),
            ).unwrap()
            # no exception, code exists
            if exists_ok:
                return existing_code
            else:
                raise SyftException(
                    public_message="UserCode with this code already exists"
                )
        except NotFoundException:
            pass

        code = submit_code.to(UserCode, context=context)
        result = self._post_user_code_transform_ops(context, code)

        if result.is_err():
            # if the validation fails, we should remove the user code status
            # and code version to prevent dangling status
            root_context = AuthedServiceContext(
                credentials=context.server.verify_key, server=context.server
            )

            if code.status_link is not None:
                _ = context.server.services.user_code_status.remove(
                    root_context, code.status_link.object_uid
                )

            # result.unwrap() will raise any exceptions from post_user_code_transform_ops
            result.unwrap()

        return self.stash.set(context.credentials, code).unwrap()

    @service_method(
        path="code.update",
        name="update",
        roles=ADMIN_ROLE_LEVEL,
        autosplat=["code_update"],
        unwrap_on_success=False,
    )
    def update(
        self,
        context: AuthedServiceContext,
        code_update: UserCodeUpdate,
    ) -> SyftSuccess:
        updated_code = self.stash.update(context.credentials, code_update).unwrap()
        return SyftSuccess(message="UserCode updated successfully", value=updated_code)

    @service_method(
        path="code.delete",
        name="delete",
        roles=ADMIN_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def delete(self, context: AuthedServiceContext, uid: UID) -> SyftSuccess:
        """Delete User Code"""
        self.stash.delete_by_uid(context.credentials, uid).unwrap()
        return SyftSuccess(message=f"User Code {uid} deleted", value=uid)

    @service_method(
        path="code.get_by_service_func_name",
        name="get_by_service_func_name",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_service_name(
        self, context: AuthedServiceContext, service_func_name: str
    ) -> list[UserCode]:
        return self.stash.get_by_service_func_name(
            context.credentials, service_func_name=service_func_name
        ).unwrap()

    # TODO: Add usercode errors
    @as_result(SyftException)
    def _post_user_code_transform_ops(
        self,
        context: AuthedServiceContext,
        user_code: UserCode,
    ) -> UserCode:
        if user_code.output_readers is None:
            raise SyftException(
                public_message=f"there is no verified output readers for {user_code}"
            )
        if user_code.input_owner_verify_keys is None:
            raise SyftException(
                public_message=f"there is no verified input owners for {user_code}"
            )
        if not all(
            x in user_code.input_owner_verify_keys for x in user_code.output_readers
        ):
            raise SyftException(
                public_message="outputs can only be distributed to input owners"
            )
        context.server.services.syft_worker_pool._get_worker_pool(
            context,
            pool_name=user_code.worker_pool_name,
        )

        # Create a code history
        context.server.services.code_history.submit_version(
            context=context, code=user_code
        )

        return user_code

    @as_result(SyftException)
    def _request_code_execution(
        self,
        context: AuthedServiceContext,
        user_code: UserCode,
        reason: str | None = "",
    ) -> Request:
        # Cannot make multiple requests for the same code
        # FIX: Change requestservice result type
        existing_requests = context.server.services.request.get_by_usercode_id(
            context, user_code.id
        )

        if len(existing_requests) > 0:
            raise SyftException(
                public_message=(
                    f"Request {existing_requests[0].id} already exists for this UserCode."
                    f" Please use the existing request, or submit a new UserCode to create a new request."
                )
            )

        # Users that have access to the output also have access to the code item
        if user_code.output_readers is not None:
            self.stash.add_permissions(
                [
                    ActionObjectPermission(user_code.id, ActionPermission.READ, x)
                    for x in user_code.output_readers
                ]
            )

        code_link = LinkedObject.from_obj(user_code, server_uid=context.server.id)

        # Requests made on low side are synced, and have their status computed instead of set manually.
        if user_code.is_l0_deployment:
            status_change = SyncedUserCodeStatusChange(
                value=UserCodeStatus.APPROVED,
                linked_obj=user_code.status_link,
                linked_user_code=code_link,
            )
        else:
            status_change = UserCodeStatusChange(
                value=UserCodeStatus.APPROVED,
                linked_obj=user_code.status_link,
                linked_user_code=code_link,
            )
        changes = [status_change]

        request = SubmitRequest(changes=changes)
        result = context.server.services.request.submit(
            context=context, request=request, reason=reason
        )

        return result

    @as_result(SyftException, NotFoundException, StashException)
    def _get_or_submit_user_code(
        self,
        context: AuthedServiceContext,
        code: SubmitUserCode | UserCode,
    ) -> UserCode:
        """
        - If the code is a UserCode, check if it exists and return
        - If the code is a SubmitUserCode and the same code hash exists, return the existing code
        - If the code is a SubmitUserCode and the code hash does not exist, submit the code
        """
        if isinstance(code, UserCode):
            return self.stash.get_by_uid(context.credentials, code.id).unwrap()
        else:  # code: SubmitUserCode
            # Submit new UserCode, or get existing UserCode with the same code hash
            # TODO: Why is this tagged as unreachable?
            return self._submit(context, code, exists_ok=True).unwrap()  # type: ignore[unreachable]

    @service_method(
        path="code.request_code_execution",
        name="request_code_execution",
        roles=GUEST_ROLE_LEVEL,
    )
    def request_code_execution(
        self,
        context: AuthedServiceContext,
        code: SubmitUserCode | UserCode,
        reason: str | None = "",
    ) -> Request:
        """Request Code execution on user code"""
        user_code = self._get_or_submit_user_code(context, code).unwrap()

        result = self._request_code_execution(
            context,
            user_code,
            reason,
        ).unwrap()

        return result

    @service_method(path="code.get_all", name="get_all", roles=GUEST_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> list[UserCode]:
        """Get a Dataset"""
        return self.stash.get_all(context.credentials).unwrap()

    @service_method(
        path="code.get_by_id", name="get_by_id", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_by_uid(self, context: AuthedServiceContext, uid: UID) -> UserCode:
        """Get a User Code Item"""
        user_code = self.stash.get_by_uid(context.credentials, uid=uid).unwrap()
        if user_code and user_code.input_policy_state and context.server is not None:
            # TODO replace with LinkedObject Context
            user_code.server_uid = context.server.id
        return user_code

    @service_method(
        path="code.get_all_for_user",
        name="get_all_for_user",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_for_user(self, context: AuthedServiceContext) -> list[UserCode]:
        """Get All User Code Items for User's VerifyKey"""
        # TODO: replace with incoming user context and key
        return self.stash.get_all(context.credentials).unwrap()

    def update_code_state(
        self, context: AuthedServiceContext, code_item: UserCode
    ) -> UserCode:
        context = context.as_root_context()
        return self.stash.update(context.credentials, code_item).unwrap()

    @as_result(SyftException)
    def load_user_code(self, context: AuthedServiceContext) -> None:
        user_code_items = self.stash.get_all(credentials=context.credentials).unwrap()
        load_approved_policy_code(user_code_items=user_code_items, context=context)

    # FIX: Exceptions etc
    def is_execution_allowed(
        self,
        code: UserCode,
        context: AuthedServiceContext,
        output_policy: OutputPolicy | None,
    ) -> IsExecutionAllowedEnum:
        status = code.get_status(context).unwrap()
        if not status.get_is_approved(context):
            return IsExecutionAllowedEnum.NOT_APPROVED
        elif self.has_code_permission(code, context) is HasCodePermissionEnum.DENIED:
            # TODO: Check enum above
            return IsExecutionAllowedEnum.NO_PERMISSION
        elif not code.is_output_policy_approved(context):
            return IsExecutionAllowedEnum.OUTPUT_POLICY_NOT_APPROVED

        if output_policy is None:
            return IsExecutionAllowedEnum.OUTPUT_POLICY_NONE

        try:
            output_policy.is_valid(context)
        except Exception:
            return IsExecutionAllowedEnum.INVALID_OUTPUT_POLICY

        return IsExecutionAllowedEnum.ALLOWED

    def is_execution_on_owned_args_allowed(self, context: AuthedServiceContext) -> bool:
        if context.role == ServiceRole.ADMIN:
            return True
        current_user = context.server.services.user.get_current_user(context=context)
        return current_user.mock_execution_permission

    def keep_owned_kwargs(
        self, kwargs: dict[str, Any], context: AuthedServiceContext
    ) -> dict[str, Any]:
        """Return only the kwargs that are owned by the user"""
        mock_kwargs = {}
        for k, v in kwargs.items():
            if isinstance(v, UID):
                # Jobs have UID kwargs instead of ActionObject
                try:
                    v = context.server.services.action.get(context, uid=v)
                except Exception:  # nosec: we are skipping when dont find it
                    pass
            if (
                isinstance(v, ActionObject)
                and v.syft_client_verify_key == context.credentials
            ):
                mock_kwargs[k] = v
        return mock_kwargs

    def is_execution_on_owned_args(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        passed_kwargs: dict[str, Any],
    ) -> bool:
        # Check if all kwargs are owned by the user
        all_kwargs_are_owned = len(
            self.keep_owned_kwargs(passed_kwargs, context)
        ) == len(passed_kwargs)

        if not all_kwargs_are_owned:
            return False

        # Check if the kwargs match the code signature
        try:
            code = self.stash.get_by_uid(context.credentials, user_code_id).unwrap()
        except SyftException:
            return False

        # Skip the datasite and context kwargs, they are passed by the backend
        code_kwargs = set(code.signature.parameters.keys()) - {"datasite", "context"}

        passed_kwarg_keys = set(passed_kwargs.keys())
        return passed_kwarg_keys == code_kwargs

    @service_method(path="code.call", name="call", roles=GUEST_ROLE_LEVEL)
    def call(
        self, context: AuthedServiceContext, uid: UID, **kwargs: Any
    ) -> ActionObject:
        """Call a User Code Function"""
        kwargs.pop("result_id", None)
        return self._call(context, uid, **kwargs).unwrap()

    def valid_worker_pool_for_context(
        self, context: AuthedServiceContext, user_code: UserCode
    ) -> bool:
        """This is a temporary fix that is needed until every function is always just ran as job"""
        # relative
        from ...server.server import get_default_worker_pool_name

        has_custom_worker_pool = (
            user_code.worker_pool_name is not None
        ) and user_code.worker_pool_name != get_default_worker_pool_name()
        if has_custom_worker_pool and context.is_blocking_api_call:
            return False
        else:
            return True

    @as_result(SyftException)
    def _call(
        self,
        context: AuthedServiceContext,
        uid: UID,
        result_id: UID | None = None,
        **kwargs: Any,
    ) -> ActionObject:
        """Call a User Code Function"""
        code: UserCode = self.stash.get_by_uid(context.credentials, uid=uid).unwrap()

        # Set Permissions
        if self.is_execution_on_owned_args(context, uid, kwargs):
            if self.is_execution_on_owned_args_allowed(context):
                # handles the case: if we have 1 or more owned args and execution permission
                # handles the case: if we have 0 owned args and execution permission
                context.has_execute_permissions = True
            elif len(kwargs) == 0:
                # handles the case: if we have 0 owned args and execution permission
                pass
            else:
                raise SyftException(
                    public_message="You do not have the permissions for mock execution, please contact the admin"
                )

        override_execution_permission = (
            context.has_execute_permissions or context.role == ServiceRole.ADMIN
        )

        # Override permissions bypasses the cache, since we do not check in/out policies
        skip_fill_cache = override_execution_permission
        # We do not read from output policy cache if there are mock arguments
        skip_read_cache = len(self.keep_owned_kwargs(kwargs, context)) > 0

        # Extract ids from kwargs
        kwarg2id = map_kwargs_to_id(kwargs)

        input_policy = code.get_input_policy(context)
        output_policy = code.get_output_policy(context)

        # Check output policy
        if not override_execution_permission:
            output_history = code.get_output_history(context=context).unwrap()

            is_execution_allowed = self.is_execution_allowed(
                code=code,
                context=context,
                output_policy=output_policy,
            )

            if (
                is_execution_allowed is not IsExecutionAllowedEnum.ALLOWED
                or context.is_l0_lowside
            ):
                # We check output policy only in l2 deployment.
                # code is from low side (L0 setup)
                status = code.get_status(context).unwrap()

                if (
                    context.server_allows_execution_for_ds
                    and not status.get_is_approved(context)
                ):
                    raise SyftException(
                        public_message=status.get_status_message_l2(context)
                    )

                output_policy_is_valid = False
                try:
                    if output_policy:
                        output_policy_is_valid = output_policy.is_valid(context)
                except SyftException:
                    pass

                # if you cant run it or the results are being sycned from l0
                # lets have a look at the output history and possibly return that
                if not output_policy_is_valid or code.is_l0_deployment:
                    if len(output_history) > 0 and not skip_read_cache:
                        last_executed_output = output_history[-1]
                        # Check if the inputs of the last executed output match
                        # against the current input
                        if (
                            input_policy is not None
                            and not last_executed_output.check_input_ids(
                                kwargs=kwarg2id
                            )
                        ):
                            inp_policy_validation = input_policy.is_valid(
                                context,
                                usr_input_kwargs=kwarg2id,
                            )

                            if not inp_policy_validation:
                                raise SyftException(
                                    # TODO: Print what's inside
                                    public_message=InputPolicyValidEnum.INVALID
                                )

                        outputs = resolve_outputs(
                            context=context,
                            output_ids=last_executed_output.output_ids,
                        ).unwrap()

                        if outputs:
                            outputs = delist_if_single(outputs)

                        if code.is_l2_deployment:
                            # Skip output policy warning in L0 setup;
                            # admin overrides policy checks.
                            output_policy_message = (
                                "Your result has been fetched from output_history, "
                                "because your OutputPolicy is no longer valid."
                            )
                            context.add_warning(output_policy_message)
                        return outputs  # type: ignore

                raise SyftException(public_message=is_execution_allowed.value)

        # Execute the code item
        if not self.valid_worker_pool_for_context(context, code):
            raise SyftException(
                public_message="You tried to run a syft function attached to a worker pool in blocking mode,"
                "which is currently not supported. Run your function with `blocking=False` to run"
                " as a job on your worker pool."
            )
        action_obj = context.server.services.action._user_code_execute(
            context, code, kwarg2id, result_id
        ).unwrap()

        result = context.server.services.action.set_result_to_store(
            action_obj, context, code.get_output_policy(context)
        ).unwrap()

        # Apply Output Policy to the results and update the OutputPolicyState

        # this currently only works for nested syft_functions
        # and admins executing on high side (TODO, decide if we want to increment counter)
        # always store_execution_output on l0 setup
        is_l0_request = context.role == ServiceRole.ADMIN and code.is_l0_deployment

        if not skip_fill_cache and output_policy is not None or is_l0_request:
            code.store_execution_output(
                context=context,
                outputs=result,
                job_id=context.job_id,
                input_ids=kwarg2id,
            ).unwrap()

        has_result_read_permission = context.extra_kwargs.get(
            "has_result_read_permission", False
        )

        # TODO: Just to fix the issue with the current implementation
        if context.role == ServiceRole.ADMIN:
            has_result_read_permission = True

        if isinstance(result, TwinObject):
            if has_result_read_permission:
                return result.private
            else:
                return result.mock
        elif result.is_mock:  # type: ignore[unreachable]
            return result  # type: ignore[return-value]
        # TODO: Check this part after error handling PR
        elif result.syft_action_data_type is Err:
            # result contains the error but the request was handled correctly
            return result
        elif has_result_read_permission:
            return result
        else:
            return result.as_empty()

    def has_code_permission(
        self, code_item: UserCode, context: AuthedServiceContext
    ) -> HasCodePermissionEnum:
        if not (
            context.credentials == context.server.verify_key
            or context.credentials == code_item.user_verify_key
        ):
            return HasCodePermissionEnum.DENIED
        return HasCodePermissionEnum.ACCEPTED

    @service_method(
        path="code.store_execution_output",
        name="store_execution_output",
        roles=GUEST_ROLE_LEVEL,
    )
    def store_execution_output(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        outputs: Any,
        input_ids: dict[str, UID] | None = None,
        job_id: UID | None = None,
    ) -> ExecutionOutput:
        code: UserCode = self.stash.get_by_uid(
            context.credentials, user_code_id
        ).unwrap()

        is_admin = context.role == ServiceRole.ADMIN

        if (
            not code.get_status(context).unwrap().get_is_approved(context)
            and not is_admin
        ):
            raise SyftException(public_message="This UserCode is not approved")

        return code.store_execution_output(
            context=context,
            outputs=outputs,
            job_id=job_id,
            input_ids=input_ids,
        ).unwrap()


@as_result(SyftException)
def resolve_outputs(
    context: AuthedServiceContext,
    output_ids: list[UID],
) -> list[ActionObject] | None:
    # relative
    from ...service.action.action_object import TwinMode

    if isinstance(output_ids, list):
        if len(output_ids) == 0:
            return None

        outputs = []
        for output_id in output_ids:
            if context.server is not None:
                output = context.server.services.action.get(
                    context, uid=output_id, twin_mode=TwinMode.PRIVATE
                )
                outputs.append(output)
        return outputs
    else:
        raise SyftException(public_message="Cannot resolve type of output_ids")


T = TypeVar("T")


def delist_if_single(result: list[T]) -> T | list[T]:
    if len(result) == 1:
        return result[0]
    return result


def map_kwargs_to_id(kwargs: dict[str, Any]) -> dict[str, Any]:
    # relative
    from ...types.twin_object import TwinObject
    from ..action.action_object import ActionObject
    from ..dataset.dataset import Asset

    filtered_kwargs = {}
    for k, v in kwargs.items():
        value = v
        if isinstance(v, ActionObject):
            value = v.id
        if isinstance(v, TwinObject):
            value = v.id
        if isinstance(v, Asset):
            value = v.action_id

        if not isinstance(value, UID):
            raise Exception(f"Input {k} must have a UID not {type(v)}")
        filtered_kwargs[k] = value
    return filtered_kwargs


TYPE_TO_SERVICE[UserCode] = UserCodeService
SERVICE_TO_TYPES[UserCodeService].update({UserCode})



# stdlib
import contextlib

# third party
import docker
import pydantic

# relative
from ...custom_worker.config import PrebuiltWorkerConfig
from ...custom_worker.config import WorkerConfig
from ...custom_worker.k8s import IN_KUBERNETES
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...types.datetime import DateTime
from ...types.dicttuple import DictTuple
from ...types.errors import SyftException
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .image_registry import SyftImageRegistry
from .utils import image_build
from .utils import image_push
from .worker_image import SyftWorkerImage
from .worker_image import SyftWorkerImageIdentifier
from .worker_image_stash import SyftWorkerImageStash


@serializable(canonical_name="SyftWorkerImageService", version=1)
class SyftWorkerImageService(AbstractService):
    stash: SyftWorkerImageStash

    def __init__(self, store: DBManager) -> None:
        self.stash = SyftWorkerImageStash(store=store)

    @service_method(
        path="worker_image.submit",
        name="submit",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def submit(
        self, context: AuthedServiceContext, worker_config: WorkerConfig
    ) -> SyftSuccess:
        image_identifier: SyftWorkerImageIdentifier | None = None
        if isinstance(worker_config, PrebuiltWorkerConfig):
            try:
                image_identifier = SyftWorkerImageIdentifier.from_str(worker_config.tag)
            except Exception:
                raise SyftException(
                    public_message=(
                        f"Invalid Docker image name: {worker_config.tag}.\n"
                        + "Please specify the image name in this format <registry>/<repo>:<tag>."
                    )
                )
        worker_image = SyftWorkerImage(
            config=worker_config,
            created_by=context.credentials,
            image_identifier=image_identifier,
        )

        # TODO: I think this was working in python mode due to a bug because
        # it wasn't saying it was duplicate
        # why can we only have a prebuilt or a non prebuilt with the same tag?
        # bigquery uses prebuilt but we need to build and then test that prebuilt works
        # so we kind of need to use one then the other and have it pull from the first
        stored_image = self.stash.set(
            context.credentials, worker_image, ignore_duplicates=True
        ).unwrap()

        return SyftSuccess(
            message=f"Dockerfile ID: {worker_image.id} successfully submitted.",
            value=stored_image,
        )

    @service_method(
        path="worker_image.build",
        name="build",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def build(
        self,
        context: AuthedServiceContext,
        image_uid: UID,
        tag: str,
        registry_uid: UID | None = None,
        pull_image: bool = True,
        force_build: bool = False,
    ) -> SyftSuccess:
        registry: SyftImageRegistry | None = None

        if IN_KUBERNETES and registry_uid is None:
            raise SyftException(
                public_message="Registry UID is required in Kubernetes mode."
            )

        worker_image = self.stash.get_by_uid(
            credentials=context.credentials, uid=image_uid
        ).unwrap()
        if registry_uid:
            # get registry from image registry service
            registry = context.server.services.syft_image_registry.get_by_id(
                context, registry_uid
            )

        try:
            if registry:
                image_identifier = SyftWorkerImageIdentifier.with_registry(
                    tag=tag, registry=registry
                )
            else:
                image_identifier = SyftWorkerImageIdentifier.from_str(tag=tag)
        except pydantic.ValidationError as e:
            raise SyftException(public_message=f"Failed to create tag: {e}")

        # if image is already built and identifier is unchanged, return an error
        if (
            worker_image.built_at
            and worker_image.image_identifier
            and worker_image.image_identifier.full_name_with_tag
            == image_identifier.full_name_with_tag
            and not force_build
        ):
            raise SyftException(
                public_message=f"Image ID: {image_uid} is already built"
            )

        worker_image.image_identifier = image_identifier
        result = None

        if not context.server.in_memory_workers:
            build_result = image_build(worker_image, pull=pull_image).unwrap()

            worker_image.image_hash = build_result.image_hash
            worker_image.built_at = DateTime.now()

            result = SyftSuccess(
                message=f"Build for Worker ID: {worker_image.id} succeeded.\n{build_result.logs}"
            )
        else:
            result = SyftSuccess(
                message="Image building skipped, since using in-memory workers."
            )

        self.stash.update(context.credentials, obj=worker_image).unwrap()
        return result

    @service_method(
        path="worker_image.push",
        name="push",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def push(
        self,
        context: AuthedServiceContext,
        image_uid: UID,
        username: str | None = None,
        password: str | None = None,
    ) -> SyftSuccess:
        worker_image = self.stash.get_by_uid(
            credentials=context.credentials, uid=image_uid
        ).unwrap()

        if not worker_image.is_built:
            raise SyftException(
                public_message=f"Image ID: {worker_image.id} is not built yet."
            )
        elif (
            worker_image.image_identifier is None
            or worker_image.image_identifier.registry_host == ""
        ):
            raise SyftException(
                public_message=f"Image ID: {worker_image.id} does not have a valid registry host."
            )

        image_push(image=worker_image, username=username, password=password).unwrap()

        return SyftSuccess(
            message=f'Pushed Image ID: {worker_image.id} to "{worker_image.image_identifier.full_name_with_tag}".'
        )

    @service_method(
        path="worker_image.get_all",
        name="get_all",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all(self, context: AuthedServiceContext) -> DictTuple[str, SyftWorkerImage]:
        """
        One image one docker file for now
        """
        images = self.stash.get_all(credentials=context.credentials).unwrap()
        return DictTuple({image.id.to_string(): image for image in images})

    @service_method(
        path="worker_image.remove",
        name="remove",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def remove(self, context: AuthedServiceContext, uid: UID) -> SyftSuccess:
        #  Delete Docker image given image tag
        image = self.stash.get_by_uid(credentials=context.credentials, uid=uid).unwrap()

        if context.server.in_memory_workers:
            pass
        elif IN_KUBERNETES:
            # TODO: Implement image deletion in kubernetes
            raise SyftException(
                public_message="Image Deletion is not yet implemented in Kubernetes !!"
            )
        elif image and image.image_identifier:
            try:
                full_tag: str = image.image_identifier.full_name_with_tag
                with contextlib.closing(docker.from_env()) as client:
                    client.images.remove(image=full_tag)
            except docker.errors.ImageNotFound:
                raise SyftException(public_message=f"Image Tag: {full_tag} not found.")
            except Exception as e:
                raise SyftException(
                    public_message=f"Failed to delete Image Tag: {full_tag}. Error: {e}"
                )

        self.stash.delete_by_uid(credentials=context.credentials, uid=uid).unwrap()
        return SyftSuccess(message=f"Image ID: {uid} deleted successfully.")

    @service_method(
        path="worker_image.get_by_uid",
        name="get_by_uid",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_by_uid(self, context: AuthedServiceContext, uid: UID) -> SyftWorkerImage:
        return self.stash.get_by_uid(credentials=context.credentials, uid=uid).unwrap()

    @service_method(
        path="worker_image.get_by_config",
        name="get_by_config",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_by_config(
        self, context: AuthedServiceContext, worker_config: WorkerConfig
    ) -> SyftWorkerImage:
        return self.stash.get_by_worker_config(
            credentials=context.credentials, config=worker_config
        ).unwrap()



# stdlib

# third party

# third party
from sqlalchemy.orm import Session

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.db.stash import with_session
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from ...types.uid import UID
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from .worker_pool import ConsumerState
from .worker_pool import SyftWorker


@serializable(canonical_name="WorkerSQLStash", version=1)
class WorkerStash(ObjectStash[SyftWorker]):
    @as_result(StashException)
    @with_session
    def set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftWorker,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
        session: Session = None,
        skip_check_type: bool = False,
    ) -> SyftWorker:
        # By default all worker pools have all read permission
        add_permissions = [] if add_permissions is None else add_permissions
        add_permissions.append(
            ActionObjectPermission(uid=obj.id, permission=ActionPermission.ALL_READ)
        )
        return (
            super()
            .set(
                credentials,
                obj,
                add_permissions=add_permissions,
                ignore_duplicates=ignore_duplicates,
                add_storage_permission=add_storage_permission,
                session=session,
            )
            .unwrap()
        )

    @as_result(StashException, NotFoundException)
    def update_consumer_state(
        self, credentials: SyftVerifyKey, worker_uid: UID, consumer_state: ConsumerState
    ) -> SyftWorker:
        worker = self.get_by_uid(credentials=credentials, uid=worker_uid).unwrap()
        worker.consumer_state = consumer_state
        return self.update(credentials=credentials, obj=worker).unwrap()



# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ..service import AbstractService


@serializable()
class RemoteProfile(SyftObject):
    __canonical_name__ = "RemoteConfig"
    __version__ = SYFT_OBJECT_VERSION_1


@serializable()
class AzureRemoteProfile(RemoteProfile):
    __canonical_name__ = "AzureRemoteConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    profile_name: str  # used by seaweedfs
    account_name: str
    account_key: str
    container_name: str


@serializable(canonical_name="RemoteProfileSQLStash", version=1)
class RemoteProfileStash(ObjectStash[RemoteProfile]):
    pass


@serializable(canonical_name="RemoteProfileService", version=1)
class RemoteProfileService(AbstractService):
    stash: RemoteProfileStash

    def __init__(self, store: DBManager) -> None:
        self.stash = RemoteProfileStash(store=store)



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from .user_code import UserCode


@serializable(canonical_name="UserCodeSQLStash", version=1)
class UserCodeStash(ObjectStash[UserCode]):
    @as_result(StashException, NotFoundException)
    def get_by_code_hash(self, credentials: SyftVerifyKey, code_hash: str) -> UserCode:
        return self.get_one(
            credentials=credentials,
            filters={"code_hash": code_hash},
        ).unwrap()

    @as_result(StashException)
    def get_by_service_func_name(
        self, credentials: SyftVerifyKey, service_func_name: str
    ) -> list[UserCode]:
        return self.get_all(
            credentials=credentials,
            filters={"service_func_name": service_func_name},
        ).unwrap()



# stdlib
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import field
import json
import random
import re
import secrets
import textwrap
from typing import Any

# relative
from ... import test_settings
from .email_helpers import TestUser

from ...client.client import SyftClient  # noqa

dataset_1 = test_settings.get("dataset_1", default="dataset_1")
dataset_2 = test_settings.get("dataset_2", default="dataset_2")
table_1 = test_settings.get("table_1", default="table_1")
table_2 = test_settings.get("table_2", default="table_2")
table_1_col_id = test_settings.get("table_1_col_id", default="table_id")
table_1_col_score = test_settings.get("table_1_col_score", default="colname")
table_2_col_id = test_settings.get("table_2_col_id", default="table_id")
table_2_col_score = test_settings.get("table_2_col_score", default="colname")


@dataclass
class TestJob:
    user_email: str
    func_name: str
    query: str
    job_type: str
    settings: dict  # make a type so we can rely on attributes
    should_succeed: bool
    should_submit: bool = True
    code_path: str | None = field(default=None)
    admin_reviewed: bool = False
    result_as_expected: bool | None = None

    _client_cache: SyftClient | None = field(default=None, repr=False, init=False)

    @property
    def is_submitted(self) -> bool:
        return self.code_path is not None

    @property
    def client(self):
        return self._client_cache

    @client.setter
    def client(self, client):
        self._client_cache = client

    def to_dict(self) -> dict:
        output = {}
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue
            output[k] = v
        return output

    def __iter__(self):
        for key, val in self.to_dict().items():
            if key.startswith("_"):
                yield key, val

    def __getitem__(self, key):
        if key.startswith("_"):
            return None
        return self.to_dict()[key]

    @property
    def code_method(self) -> None | Callable:
        try:
            return getattr(self.client.code, self.func_name, None)
        except Exception as e:
            print(f"Cant find code method. {e}")
        return None


def make_query(settings: dict) -> str:
    query = f"""
    SELECT {settings['groupby_col']}, AVG({settings['score_col']}) AS average_score
    FROM {settings['dataset']}.{settings['table']}
    GROUP BY {settings['groupby_col']}
    LIMIT {settings['limit']}""".strip()  # nosec: B608

    return textwrap.dedent(query)


def create_simple_query_job(user: TestUser) -> TestJob:
    job_type = "simple_query"
    func_name = f"{job_type}_{secrets.token_hex(3)}"

    dataset = random.choice([dataset_1, dataset_2])  # nosec: B311
    table, groupby_col, score_col = random.choice(  # nosec: B311
        [
            (table_1, table_1_col_id, table_1_col_score),
            (table_2, table_2_col_id, table_2_col_score),
        ]
    )
    limit = random.randint(1, 1_000_000)  # nosec: B311

    settings = {
        "dataset": dataset,
        "table": table,
        "groupby_col": groupby_col,
        "score_col": score_col,
        "limit": limit,
    }
    query = make_query(settings)

    result = TestJob(
        user_email=user.email,
        func_name=func_name,
        query=query,
        job_type=job_type,
        settings=settings,
        should_succeed=True,
    )

    result.client = user.client
    return result


def create_wrong_asset_query(user: TestUser) -> TestJob:
    job_type = "wrong_asset_query"
    func_name = f"{job_type}_{secrets.token_hex(3)}"

    valid_job = create_simple_query_job(user)
    settings = valid_job.settings
    corrupted_asset = random.choice(["dataset", "table"])  # nosec: B311
    settings[corrupted_asset] = "wrong_asset"
    query = make_query(settings)

    result = TestJob(
        user_email=user.email,
        func_name=func_name,
        query=query,
        job_type=job_type,
        settings=settings,
        should_succeed=False,
    )

    result.client = user.client
    return result


def create_wrong_syntax_query(user: TestUser) -> TestJob:
    job_type = "wrong_syntax_query"
    func_name = f"{job_type}_{secrets.token_hex(3)}"

    query = "SELECT * FROM table INCORRECT SYNTAX"

    result = TestJob(
        user_email=user.email,
        func_name=func_name,
        query=query,
        job_type=job_type,
        settings={},
        should_succeed=False,
    )

    result.client = user.client
    return result


def create_long_query_job(user: TestUser) -> TestJob:
    job_type = "job_too_much_text"
    func_name = f"{job_type}_{secrets.token_hex(3)}"

    query = "a" * 1_000

    result = TestJob(
        user_email=user.email,
        func_name=func_name,
        query=query,
        job_type=job_type,
        settings={},
        should_succeed=False,
    )

    result.client = user.client
    return result


def create_query_long_name(user: TestUser) -> TestJob:
    job_type = "job_long_name"
    func_name = f"{job_type}_{secrets.token_hex(3)}"

    job = create_simple_query_job(user)

    job.job_type = job_type
    job.func_name = func_name + "a" * 1_000

    return job


def create_job_funcname_xss(user: TestUser) -> TestJob:
    job_type = "job_funcname_xss"
    func_name = f"{job_type}_{secrets.token_hex(3)}"
    func_name += "<script>alert('XSS in funcname')</script>"

    job = create_simple_query_job(user)
    job.job_type = job_type
    job.func_name = func_name
    job.should_submit = False
    return job


def get_request_for_job_info(requests, job):
    job_requests = [r for r in requests if r.code.service_func_name == job.func_name]
    if len(job_requests) != 1:
        raise Exception(f"Too many or too few requests: {job} in requests: {requests}")
    return job_requests[0]


def create_job_query_xss(user: TestUser) -> TestJob:
    job_type = "job_query_xss"
    func_name = f"{job_type}_{secrets.token_hex(3)}"

    job = create_simple_query_job(user)
    job.job_type = job_type
    job.func_name = func_name
    job.query += "<script>alert('XSS in query')</script>"
    job.should_succeed = False

    return job


def create_job_many_columns(user: TestUser) -> TestJob:
    job_type = "job_many_columns"
    func_name = f"{job_type}_{secrets.token_hex(3)}"

    job = create_simple_query_job(user)
    job.job_type = job_type
    job.func_name = func_name
    settings = job.settings
    job.settings["num_extra_cols"] = random.randint(100, 1000)  # nosec: B311

    new_columns_string = ", ".join(
        f"{settings['score_col']} as col_{i}" for i in range(settings["num_extra_cols"])
    )

    job.query = f"""
    SELECT {settings['groupby_col']}, AVG({settings['score_col']}) AS average_score, {new_columns_string}
    FROM {settings['dataset']}.{settings['table']}
    GROUP BY {settings['groupby_col']}
    LIMIT {settings['limit']}""".strip()  # nosec: B608

    return job


def create_random_job(user: TestUser) -> TestJob:
    job_func = random.choice(create_job_functions)  # nosec: B311
    return job_func(user)


def create_jobs(users: list[TestUser], total_jobs: int = 10) -> list[TestJob]:
    jobs = []
    num_users = len(users)
    user_index = 0
    each_count = 0
    # keep making jobs until we have enough
    while len(jobs) < total_jobs:
        # if we havent used each job type yet keep getting the next one
        if each_count < len(create_job_functions):
            job_func = create_job_functions[each_count]
            each_count += 1
        else:
            # otherwise lets get a random one
            job_func = create_random_job
        # use the current index of user
        jobs.append(job_func(users[user_index]))

        # only go as high as the last user index
        if user_index < num_users - 1:
            user_index += 1
        else:
            # reset back to the first user
            user_index = 0

    # in case we stuffed up
    if len(jobs) > total_jobs:
        jobs = jobs[:total_jobs]
    return jobs


def submit_job(job: TestJob) -> tuple[Any, str]:
    client = job.client
    response = client.api.services.bigquery.submit_query(
        func_name=job.func_name, query=job.query
    )
    job.code_path = extract_code_path(response)
    return response


def extract_code_path(response) -> str | None:
    pattern = r"client\.code\.(\w+)\(\)"
    match = re.search(pattern, str(response))
    if match:
        extracted_code = match.group(1)
        return extracted_code
    return None


def approve_by_running(request):
    job = request.code(blocking=False)
    result = job.wait()
    print("got result of type", type(result), "bool", bool(result))
    # got result of type <class 'syft.service.action.pandas.PandasDataFrameObject'> bool False
    # assert result won't work unless we know what type is coming back
    job_info = job.info(result=True)
    # need force when running multiple times
    # todo check and dont run if its already done
    response = request.deposit_result(job_info, approve=True, force=True)
    return response


def get_job_emails(jobs, client, email_server):
    all_requests = client.requests
    res = {}
    for job in jobs:
        request = get_request_for_job_info(all_requests, job)
        emails = email_server.get_emails_for_user(request.requesting_user_email)
        res[request.requesting_user_email] = emails
    return res


def resolve_request(request):
    service_func_name = request.code.service_func_name
    if service_func_name.startswith("simple_query"):
        request.approve()  # approve because it is good
    if service_func_name.startswith("wrong_asset_query"):
        request.approve()  # approve because it is bad
    if service_func_name.startswith("wrong_syntax_query"):
        request.approve()  # approve because it is bad
    if service_func_name.startswith("job_too_much_text"):
        request.deny(reason="too long, boring!")  # deny because it is bad
    if service_func_name.startswith("job_long_name"):
        request.approve()
    if service_func_name.startswith("job_funcname_xss"):
        request.deny(reason="too long, boring!")  # never reach doesnt matter
    if service_func_name.startswith("job_query_xss"):
        request.approve()  # approve because it is bad
    if service_func_name.startswith("job_many_columns"):
        request.approve()  # approve because it is bad

    return (request.id, request.status)


create_job_functions = [
    create_simple_query_job,  # quick way to increase the odds
    create_simple_query_job,
    create_simple_query_job,
    create_simple_query_job,
    create_simple_query_job,
    create_simple_query_job,
    create_wrong_syntax_query,
    create_long_query_job,
    create_query_long_name,
    create_job_funcname_xss,
    create_job_query_xss,
    create_job_many_columns,
]


def save_jobs(jobs, filepath="./jobs.json"):
    user_jobs = defaultdict(list)
    for job in jobs:
        user_jobs[job.user_email].append(job.to_dict())
    with open(filepath, "w") as f:
        f.write(json.dumps(user_jobs))


def load_jobs(users, high_client, filepath="./jobs.json"):
    data = {}
    try:
        with open(filepath) as f:
            data = json.loads(f.read())
    except Exception as e:
        print(f"cant read file: {filepath}: {e}")
        data = {}
    jobs_list = []
    for user in users:
        if user.email not in data:
            print(f"{user.email} missing from jobs")
            continue
        user_jobs = data[user.email]
        for user_job in user_jobs:
            test_job = TestJob(**user_job)
            if user._client_cache is None:
                user.client = high_client
            test_job.client = user.client
            jobs_list.append(test_job)
    return jobs_list



# stdlib
from collections import defaultdict
import logging
import socketserver

# relative
from ...serde.serializable import serializable
from ...service.context import AuthedServiceContext
from ...types.errors import SyftException
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.uid import UID
from ...util.util import get_queue_address
from ..response import SyftSuccess
from ..worker.worker_stash import WorkerStash
from .base_queue import AbstractMessageHandler
from .base_queue import QueueClient
from .base_queue import QueueClientConfig
from .base_queue import QueueConfig
from .queue import ConsumerType
from .queue_stash import QueueStash
from .zmq_consumer import ZMQConsumer
from .zmq_producer import ZMQProducer

logger = logging.getLogger(__name__)


@serializable()
class ZMQClientConfig(SyftObject, QueueClientConfig):
    __canonical_name__ = "ZMQClientConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]
    hostname: str = "127.0.0.1"
    queue_port: int | None = None
    # TODO: setting this to false until we can fix the ZMQ
    # port issue causing tests to randomly fail
    create_producer: bool = False
    n_consumers: int = 0
    consumer_service: str | None = None


@serializable(attrs=["host"], canonical_name="ZMQClient", version=1)
class ZMQClient(QueueClient):
    """ZMQ Client for creating producers and consumers."""

    producers: dict[str, ZMQProducer]
    consumers: defaultdict[str, list[ZMQConsumer]]

    def __init__(self, config: ZMQClientConfig) -> None:
        self.host = config.hostname
        self.producers = {}
        self.consumers = defaultdict(list)
        self.config = config

    @staticmethod
    def _get_free_tcp_port(host: str) -> int:
        with socketserver.TCPServer((host, 0), None) as s:
            free_port = s.server_address[1]

        return free_port

    def add_producer(
        self,
        queue_name: str,
        port: int | None = None,
        queue_stash: QueueStash | None = None,
        worker_stash: WorkerStash | None = None,
        context: AuthedServiceContext | None = None,
    ) -> ZMQProducer:
        """Add a producer of a queue.

        A queue can have at most one producer attached to it.
        """

        if port is None:
            if self.config.queue_port is None:
                self.config.queue_port = self._get_free_tcp_port(self.host)
                port = self.config.queue_port
            else:
                port = self.config.queue_port

        logger.info(
            f"Adding producer for queue: {queue_name} on: {get_queue_address(port)}"
        )
        producer = ZMQProducer(
            queue_name=queue_name,
            queue_stash=queue_stash,
            port=port,
            context=context,
            worker_stash=worker_stash,
        )
        self.producers[queue_name] = producer
        return producer

    def add_consumer(
        self,
        queue_name: str,
        message_handler: AbstractMessageHandler,
        service_name: str,
        address: str | None = None,
        worker_stash: WorkerStash | None = None,
        syft_worker_id: UID | None = None,
    ) -> ZMQConsumer:
        """Add a consumer to a queue

        A queue should have at least one producer attached to the group.

        """

        if address is None:
            address = get_queue_address(self.config.queue_port)

        consumer = ZMQConsumer(
            queue_name=queue_name,
            message_handler=message_handler,
            address=address,
            service_name=service_name,
            syft_worker_id=syft_worker_id,
            worker_stash=worker_stash,
        )
        self.consumers[queue_name].append(consumer)

        return consumer

    def send_message(
        self,
        message: bytes,
        queue_name: str,
        worker: bytes | None = None,
    ) -> SyftSuccess:
        producer = self.producers.get(queue_name)
        if producer is None:
            raise SyftException(
                public_message=f"No producer attached for queue: {queue_name}. Please add a producer for it."
            )
        try:
            producer.send(message=message, worker=worker)
        except Exception as e:
            # stdlib
            raise SyftException(
                public_message=f"Failed to send message to: {queue_name} with error: {e}"
            )
        return SyftSuccess(
            message=f"Successfully queued message to : {queue_name}",
        )

    def close(self) -> SyftSuccess:
        try:
            for consumers in self.consumers.values():
                for consumer in consumers:
                    # make sure look is stopped
                    consumer.close()

            for producer in self.producers.values():
                # make sure loop is stopped
                producer.close()
                # close existing connection.
        except Exception as e:
            raise SyftException(public_message=f"Failed to close connection: {e}")

        return SyftSuccess(message="All connections closed.")

    def purge_queue(self, queue_name: str) -> SyftSuccess:
        if queue_name not in self.producers:
            raise SyftException(
                public_message=f"No producer running for : {queue_name}"
            )

        producer = self.producers[queue_name]

        # close existing connection.
        producer.close()

        # add a new connection
        self.add_producer(queue_name=queue_name, address=producer.address)  # type: ignore

        return SyftSuccess(message=f"Queue: {queue_name} successfully purged")

    def purge_all(self) -> SyftSuccess:
        for queue_name in self.producers:
            self.purge_queue(queue_name=queue_name)

        return SyftSuccess(message="Successfully purged all queues.")


@serializable(canonical_name="ZMQQueueConfig", version=1)
class ZMQQueueConfig(QueueConfig):
    def __init__(
        self,
        client_type: type[ZMQClient] | None = None,
        client_config: ZMQClientConfig | None = None,
        consumer_type: ConsumerType = ConsumerType.Process,
    ):
        self.client_type = client_type or ZMQClient
        self.client_config: ZMQClientConfig = client_config or ZMQClientConfig()
        self.consumer_type = consumer_type



# stdlib

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...types.uid import UID
from ..code.user_code import SubmitUserCode
from ..code.user_code import UserCode
from ..context import AuthedServiceContext
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .code_history import CodeHistoriesDict
from .code_history import CodeHistory
from .code_history import CodeHistoryView
from .code_history import UsersCodeHistoriesDict
from .code_history_stash import CodeHistoryStash


@serializable(canonical_name="CodeHistoryService", version=1)
class CodeHistoryService(AbstractService):
    stash: CodeHistoryStash

    def __init__(self, store: DBManager) -> None:
        self.stash = CodeHistoryStash(store=store)

    @service_method(
        path="code_history.submit_version",
        name="submit_version",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def submit_version(
        self,
        context: AuthedServiceContext,
        code: SubmitUserCode | UserCode,
        comment: str | None = None,
    ) -> SyftSuccess:
        if isinstance(code, SubmitUserCode):
            code = context.server.services.user_code._submit(
                context=context, submit_code=code
            ).unwrap()

        try:
            code_history = self.stash.get_by_service_func_name_and_verify_key(
                credentials=context.credentials,
                service_func_name=code.service_func_name,
                user_verify_key=context.credentials,
            ).unwrap()
        except NotFoundException:
            code_history = CodeHistory(
                id=UID(),
                server_uid=context.server.id,
                user_verify_key=context.credentials,
                service_func_name=code.service_func_name,
            )
            self.stash.set(credentials=context.credentials, obj=code_history).unwrap()

        code_history.add_code(code=code, comment=comment)
        res = self.stash.update(
            credentials=context.credentials, obj=code_history
        ).unwrap()
        return SyftSuccess(message="Code version submit success", value=res)

    @service_method(
        path="code_history.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_all(self, context: AuthedServiceContext) -> list[CodeHistory]:
        """Get a Dataset"""
        return self.stash.get_all(context.credentials).unwrap()

    @service_method(
        path="code_history.get", name="get", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_code_by_uid(self, context: AuthedServiceContext, uid: UID) -> SyftSuccess:
        """Get a User Code Item"""
        return self.stash.get_by_uid(context.credentials, uid=uid).unwrap()

    @service_method(path="code_history.delete", name="delete", unwrap_on_success=False)
    def delete(self, context: AuthedServiceContext, uid: UID) -> SyftSuccess:
        res = self.stash.delete_by_uid(context.credentials, uid).unwrap()
        return SyftSuccess(message="Succesfully deleted", value=res)

    def fetch_histories_for_user(
        self, context: AuthedServiceContext, user_verify_key: SyftVerifyKey
    ) -> CodeHistoriesDict:
        if context.role in [ServiceRole.DATA_OWNER, ServiceRole.ADMIN]:
            code_histories = self.stash.get_by_verify_key(
                credentials=context.server.verify_key, user_verify_key=user_verify_key
            ).unwrap()
        else:
            code_histories = self.stash.get_by_verify_key(
                credentials=context.credentials, user_verify_key=user_verify_key
            ).unwrap()

        def get_code(uid: UID) -> UserCode:
            return context.server.services.user_code.stash.get_by_uid(
                credentials=context.server.verify_key,
                uid=uid,
            ).unwrap()

        code_versions_dict = {}

        for code_history in code_histories:
            user_code_list = [get_code(uid) for uid in code_history.user_code_history]
            code_versions = CodeHistoryView(
                user_code_history=user_code_list,
                service_func_name=code_history.service_func_name,
                comment_history=code_history.comment_history,
            )
            code_versions_dict[code_history.service_func_name] = code_versions
        return CodeHistoriesDict(code_versions=code_versions_dict)

    @service_method(
        path="code_history.get_history",
        name="get_history",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_histories_for_current_user(
        self, context: AuthedServiceContext
    ) -> CodeHistoriesDict:
        return self.fetch_histories_for_user(
            context=context, user_verify_key=context.credentials
        )

    @service_method(
        path="code_history.get_history_for_user",
        name="get_history_for_user",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_history_for_user(
        self, context: AuthedServiceContext, email: str
    ) -> CodeHistoriesDict:
        user = context.server.services.user.stash.get_by_email(
            credentials=context.credentials, email=email
        ).unwrap()
        return self.fetch_histories_for_user(
            context=context, user_verify_key=user.verify_key
        )

    @service_method(
        path="code_history.get_histories",
        name="get_histories",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_histories_group_by_user(
        self, context: AuthedServiceContext
    ) -> UsersCodeHistoriesDict:
        if context.role in [ServiceRole.DATA_OWNER, ServiceRole.ADMIN]:
            code_histories = self.stash.get_all(
                context.credentials, has_permission=True
            ).unwrap()
        else:
            code_histories = self.stash.get_all(context.credentials).unwrap()

        users = context.server.services.user.stash.get_all(context.credentials).unwrap()
        user_code_histories = UsersCodeHistoriesDict(server_uid=context.server.id)

        verify_key_2_user_email = {}
        for user in users:
            user_code_histories.user_dict[user.email] = []
            verify_key_2_user_email[user.verify_key] = user.email

        for code_history in code_histories:
            user_email = verify_key_2_user_email[code_history.user_verify_key]
            user_code_histories.user_dict[user_email].append(
                code_history.service_func_name
            )

        return user_code_histories

    @service_method(
        path="code_history.get_by_name_and_user_email",
        name="get_by_name_and_user_email",
    )
    def get_by_func_name_and_user_email(
        self,
        context: AuthedServiceContext,
        service_func_name: str,
        user_email: str,
        user_id: UID,
    ) -> list[CodeHistory]:
        user_verify_key = context.server.services.user.user_verify_key(user_email)

        filters = {
            "id": user_id,
            "email": user_email,
            "verify_key": user_verify_key,
            "service_func_name": service_func_name,
        }

        return self.stash.get_all(
            credentials=context.credentials, filters=filters
        ).unwrap()



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import StashException
from ...types.result import as_result
from .sync_state import SyncState


@serializable(canonical_name="SyncStash", version=1)
class SyncStash(ObjectStash[SyncState]):
    def __init__(self, store: DBManager) -> None:
        super().__init__(store)
        self.last_state: SyncState | None = None

    @as_result(StashException)
    def get_latest(self, credentials: SyftVerifyKey) -> SyncState | None:
        if self.last_state is not None:
            return self.last_state

        states = self.get_all(
            credentials=credentials,
            order_by="created_at",
            sort_order="desc",
            limit=1,
        ).unwrap()

        if len(states) > 0:
            return states[0]
        return None



# stdlib
from distutils.dir_util import copy_tree
from pathlib import Path

# relative
from ..orchestra import ServerHandle

COPY_SUFFIX = "-copy"


def make_copy_path(path: Path) -> str:
    return f"{path.parent}/{path.stem}{COPY_SUFFIX}{path.suffix}"


def make_original_path(copy_path: Path) -> str:
    return f"{str(copy_path)[:-len(COPY_SUFFIX)]}"


def make_copy(server: ServerHandle) -> None:
    if not server.python_server:
        print("server does not have python server, no copy made")
        return
    cfg = server.python_server.db_config
    original_dir = str(cfg.path.resolve())
    copy_dir = f"{original_dir}{COPY_SUFFIX}"
    copy_tree(original_dir, copy_dir)
    print(f"copied\n{original_dir}\nto\n{copy_dir}\n")


def restore_copy(copy_dir: str) -> None:
    copy_dir_path = Path(copy_dir)
    original_dir = make_original_path(copy_dir_path)
    copy_tree(copy_dir_path, original_dir)
    print(f"copied\n{copy_dir}\nto\n{original_dir}\n")



# stdlib
import os

# third party
from ipywidgets.widgets import Output as ipyOutput


class Output(ipyOutput):
    # This is a workaround for the issue that
    # the Output widget causes halt when running in Jupyter Notebook
    # from cli, e.g. tests.
    #
    # No-op when running in Jupyter Notebook.
    if "JPY_PARENT_PID" in os.environ:

        def __enter__(self):  # type: ignore
            pass

        def __exit__(self, *args, **kwargs):  # type: ignore
            pass



# stdlib
from abc import ABC
from abc import abstractmethod
import enum
from typing import Any
from typing import Literal

# third party
import sqlalchemy as sa
from sqlalchemy import Column
from sqlalchemy import Dialect
from sqlalchemy import Result
from sqlalchemy import Select
from sqlalchemy import Table
from sqlalchemy import func
from sqlalchemy.exc import DatabaseError
from sqlalchemy.orm import Session
from typing_extensions import Self

# relative
from ...serde.json_serde import serialize_json
from ...server.credentials import SyftVerifyKey
from ...service.action.action_permissions import ActionObjectPermission
from ...service.action.action_permissions import ActionPermission
from ...service.user.user_roles import ServiceRole
from ...types.syft_object import SyftObject
from ...types.uid import UID
from .errors import StashDBException
from .schema import PostgresBase
from .schema import SQLiteBase


class FilterOperator(enum.Enum):
    EQ = "eq"
    CONTAINS = "contains"


class Query(ABC):
    def __init__(self, object_type: type[SyftObject]) -> None:
        self.object_type: type = object_type
        self.table: Table = self._get_table(object_type)
        self.stmt: Select = self.table.select()

    @abstractmethod
    def _get_table(self, object_type: type[SyftObject]) -> Table:
        raise NotImplementedError

    @staticmethod
    def get_query_class(dialect: str | Dialect) -> "type[Query]":
        if isinstance(dialect, Dialect):
            dialect = dialect.name

        if dialect == "sqlite":
            return SQLiteQuery
        elif dialect == "postgresql":
            return PostgresQuery
        else:
            raise ValueError(f"Unsupported dialect {dialect}")

    @classmethod
    def create(cls, object_type: type[SyftObject], dialect: str | Dialect) -> "Query":
        """Create a query object for the given object type and dialect."""
        query_class = cls.get_query_class(dialect)
        return query_class(object_type)

    def execute(self, session: Session) -> Result:
        """Execute the query using the given session."""
        try:
            return session.execute(self.stmt)
        except DatabaseError as e:
            raise StashDBException.from_sqlalchemy_error(e) from e

    def with_permissions(
        self,
        credentials: SyftVerifyKey,
        role: ServiceRole,
        permission: ActionPermission = ActionPermission.READ,
    ) -> Self:
        """Add a permission check to the query.

        If the user has a role below DATA_OWNER, the query will be filtered to only include objects
        that the user has the specified permission on.

        Args:
            credentials (SyftVerifyKey): user verify key
            role (ServiceRole): role of the user
            permission (ActionPermission, optional): Type of permission to check for.
                Defaults to ActionPermission.READ.

        Returns:
            Self: The query object with the permission check applied
        """
        if role in (ServiceRole.ADMIN, ServiceRole.DATA_OWNER):
            return self

        ao_permission = ActionObjectPermission(
            uid=UID(),  # dummy uid, we just need the permission string
            credentials=credentials,
            permission=permission,
        )

        permission_clause = self._make_permissions_clause(ao_permission)
        self.stmt = self.stmt.where(permission_clause)

        return self

    def filter(self, field: str, operator: str | FilterOperator, value: Any) -> Self:
        """Add a filter to the query.

        example usage:
        Query(User).filter("name", "eq", "Alice")
        Query(User).filter("friends", "contains", "Bob")

        Args:
            field (str): Field to filter on
            operator (str): Operator to use for the filter
            value (Any): Value to filter on

        Raises:
            ValueError: If the operator is not supported

        Returns:
            Self: The query object with the filter applied
        """
        filter = self._create_filter_clause(self.table, field, operator, value)
        self.stmt = self.stmt.where(filter)
        return self

    def filter_and(self, *filters: tuple[str, str | FilterOperator, Any]) -> Self:
        """Add filters to the query using an AND clause.

        example usage:
        Query(User).filter_and(
            ("name", "eq", "Alice"),
            ("age", "eq", 30),
        )

        Args:
            field (str): Field to filter on
            operator (str): Operator to use for the filter
            value (Any): Value to filter on

        Raises:
            ValueError: If the operator is not supported

        Returns:
            Self: The query object with the filter applied
        """
        filter_clauses = [
            self._create_filter_clause(self.table, field, operator, value)
            for field, operator, value in filters
        ]

        self.stmt = self.stmt.where(sa.and_(*filter_clauses))
        return self

    def filter_or(self, *filters: tuple[str, str | FilterOperator, Any]) -> Self:
        """Add filters to the query using an OR clause.

        example usage:
        Query(User).filter_or(
            ("name", "eq", "Alice"),
            ("age", "eq", 30),
        )

        Args:
            field (str): Field to filter on
            operator (str): Operator to use for the filter
            value (Any): Value to filter on

        Raises:
            ValueError: If the operator is not supported

        Returns:
            Self: The query object with the filter applied
        """
        filter_clauses = [
            self._create_filter_clause(self.table, field, operator, value)
            for field, operator, value in filters
        ]

        self.stmt = self.stmt.where(sa.or_(*filter_clauses))
        return self

    def _create_filter_clause(
        self,
        table: Table,
        field: str,
        operator: str | FilterOperator,
        value: Any,
    ) -> sa.sql.elements.BinaryExpression:
        if isinstance(operator, str):
            try:
                operator = FilterOperator(operator.lower())
            except ValueError:
                raise ValueError(f"Filter operator {operator} not supported")

        if operator == FilterOperator.EQ:
            return self._eq_filter(table, field, value)
        elif operator == FilterOperator.CONTAINS:
            return self._contains_filter(table, field, value)

    def order_by(
        self,
        field: str | None = None,
        order: Literal["asc", "desc"] | None = None,
    ) -> Self:
        """Add an order by clause to the query, with sensible defaults if field or order is not provided.

        Args:
            field (Optional[str]): field to order by. If None, uses the default field.
            order (Optional[Literal["asc", "desc"]]): Order to use ("asc" or "desc").
                Defaults to 'asc' if field is provided and order is not, or the default order otherwise.

        Raises:
            ValueError: If the order is not "asc" or "desc"

        Returns:
            Self: The query object with the order by clause applied.
        """
        # Determine the field and order defaults if not provided
        if field is None:
            if hasattr(self.object_type, "__order_by__"):
                default_field, default_order = self.object_type.__order_by__
            else:
                default_field, default_order = "_created_at", "desc"
            field = default_field
        else:
            # If field is provided but order is not, default to 'asc'
            default_order = "asc"
        order = order or default_order

        column = self._get_column(field)

        if isinstance(column.type, sa.JSON):
            column = sa.cast(column, sa.String)

        if order.lower() == "asc":
            self.stmt = self.stmt.order_by(column.asc())

        elif order.lower() == "desc":
            self.stmt = self.stmt.order_by(column.desc())
        else:
            raise ValueError(f"Invalid sort order {order}")

        return self

    def limit(self, limit: int | None) -> Self:
        """Add a limit clause to the query."""
        if limit is None:
            return self

        if limit < 0:
            raise ValueError("Limit must be a positive integer")
        self.stmt = self.stmt.limit(limit)

        return self

    def offset(self, offset: int) -> Self:
        """Add an offset clause to the query."""
        if offset < 0:
            raise ValueError("Offset must be a positive integer")

        self.stmt = self.stmt.offset(offset)
        return self

    @abstractmethod
    def _make_permissions_clause(
        self,
        permission: ActionObjectPermission,
    ) -> sa.sql.elements.BinaryExpression:
        pass

    @abstractmethod
    def _contains_filter(
        self,
        table: Table,
        field: str,
        value: Any,
    ) -> sa.sql.elements.BinaryExpression:
        pass

    def _get_column(self, column: str) -> Column:
        if column == "id":
            return self.table.c.id
        if column == "created_date" or column == "_created_at":
            return self.table.c._created_at
        elif column == "updated_date" or column == "_updated_at":
            return self.table.c._updated_at
        elif column == "deleted_date" or column == "_deleted_at":
            return self.table.c._deleted_at

        return self.table.c.fields[column]


class SQLiteQuery(Query):
    def _make_permissions_clause(
        self,
        permission: ActionObjectPermission,
    ) -> sa.sql.elements.BinaryExpression:
        permission_string = permission.permission_string
        compound_permission_string = permission.compound_permission_string
        return sa.or_(
            self.table.c.permissions.contains(permission_string),
            self.table.c.permissions.contains(compound_permission_string),
        )

    def _get_table(self, object_type: type[SyftObject]) -> Table:
        cname = object_type.__canonical_name__
        if cname not in SQLiteBase.metadata.tables:
            raise ValueError(f"Table for {cname} not found")
        return SQLiteBase.metadata.tables[cname]

    def _contains_filter(
        self,
        table: Table,
        field: str,
        value: Any,
    ) -> sa.sql.elements.BinaryExpression:
        field_value = serialize_json(value)
        return table.c.fields[field].contains(func.json_quote(field_value))

    def _eq_filter(
        self,
        table: Table,
        field: str,
        value: Any,
    ) -> sa.sql.elements.BinaryExpression:
        if field == "id":
            return table.c.id == UID(value)

        if "." in field:
            # magic!
            field = field.split(".")  # type: ignore

        json_value = serialize_json(value)
        return table.c.fields[field] == func.json_quote(json_value)


class PostgresQuery(Query):
    def _make_permissions_clause(
        self, permission: ActionObjectPermission
    ) -> sa.sql.elements.BinaryExpression:
        permission_string = [permission.permission_string]
        compound_permission_string = [permission.compound_permission_string]
        return sa.or_(
            self.table.c.permissions.contains(permission_string),
            self.table.c.permissions.contains(compound_permission_string),
        )

    def _contains_filter(
        self,
        table: Table,
        field: str,
        value: Any,
    ) -> sa.sql.elements.BinaryExpression:
        field_value = serialize_json(value)
        col = sa.cast(table.c.fields[field], sa.Text)
        val = sa.cast(field_value, sa.Text)
        return col.contains(val)

    def _get_table(self, object_type: type[SyftObject]) -> Table:
        cname = object_type.__canonical_name__
        if cname not in PostgresBase.metadata.tables:
            raise ValueError(f"Table for {cname} not found")
        return PostgresBase.metadata.tables[cname]

    def _eq_filter(
        self,
        table: Table,
        field: str,
        value: Any,
    ) -> sa.sql.elements.BinaryExpression:
        if field == "id":
            return table.c.id == UID(value)

        if "." in field:
            # magic!
            field = field.split(".")  # type: ignore

        json_value = serialize_json(value)
        # NOTE: there might be a bug with casting everything to text
        return table.c.fields[field].astext == sa.cast(json_value, sa.Text)



# stdlib

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ..context import AuthedServiceContext
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import GUEST_ROLE_LEVEL
from .server_metadata import ServerMetadata


@serializable(canonical_name="MetadataService", version=1)
class MetadataService(AbstractService):
    def __init__(self, store: DBManager) -> None:
        pass

    @service_method(
        path="metadata.get_metadata", name="get_metadata", roles=GUEST_ROLE_LEVEL
    )
    def get_metadata(self, context: AuthedServiceContext) -> ServerMetadata:
        return context.server.metadata  # type: ignore

    @service_method(path="metadata.get_env", name="get_env", roles=GUEST_ROLE_LEVEL)
    def get_env(self, context: AuthedServiceContext) -> str:
        return context.server.packages



# stdlib
from pathlib import Path
from typing import cast

# third party
from IPython.display import display

# syft absolute
import syft as sy

# relative
from ..client.client import SyftClient
from ..service.migration.object_migration_state import MigrationData
from ..service.worker.image_identifier import SyftWorkerImageIdentifier
from ..service.worker.worker_image import SyftWorkerImage
from ..service.worker.worker_pool import WorkerPool


def upgrade_custom_workerpools(
    client: SyftClient,
    migration_data: str | Path | MigrationData,
    mode: str = "manual",
) -> None:
    """Upgrade custom workerpools to the new syft version

    Args:
        client (SyftClient): Admin client to upgrade workerpools with
        migration_data (str | Path | MigrationData): Path to migration data or MigrationData object
        mode (str, optional): if "auto" the upgrade will be done automatically. "auto" assumes
            all images and tags use Syft versioning. Defaults to "manual".

    Raises:
        ValueError: if mode is not "manual" or "auto"
    """
    print("This is a utility to upgrade workerpools to the new syft version")
    print("If an upgrade fails, it is always possible to start the workerpool manually")

    if mode not in ["manual", "auto"]:
        raise ValueError("mode must be either 'manual' or 'auto'")

    if isinstance(migration_data, str | Path):
        print("loading migration data...")
        migration_data = MigrationData.from_file(migration_data)

    # mypy does not recognize instance check for str | Path
    migration_data = cast(MigrationData, migration_data)
    worker_pools = migration_data.get_items_by_canonical_name(
        WorkerPool.__canonical_name__
    )
    num_upgraded = 0
    for pool in worker_pools:
        is_upgraded = upgrade_workerpool(client, pool, migration_data, mode)
        if is_upgraded:
            num_upgraded += 1
        print()

    print(f"Upgraded {num_upgraded} workerpools to the new syft version")
    print("Please verify your upgraded pools with `client.worker_pools`")


def upgrade_workerpool(
    client: SyftClient,
    pool: WorkerPool,
    migration_data: MigrationData,
    mode: str = "manual",
) -> bool:
    if pool.name == migration_data.default_pool_name:
        print("Skipping default pool, this pool has already been upgraded")
        return False

    print(f"Upgrading workerpool {pool.name}")

    images = migration_data.get_items_by_canonical_name(
        SyftWorkerImage.__canonical_name__
    )
    image_id = pool.image_id
    old_image: SyftWorkerImage = [img for img in images if img.id == image_id][0]

    if old_image.is_prebuilt:
        new_image = upgrade_prebuilt_image(client, old_image, mode)
    else:
        new_image = upgrade_syft_image(client, old_image, mode)

    if not new_image:
        print(f"Failed to upgrade workerpool {pool.name}, could not build new image")
        return False

    print(f"starting new pool `{pool.name}` with {pool.max_count} workers")
    try:
        result = client.api.services.worker_pool.launch(
            pool_name=pool.name,
            image_uid=new_image.id,
            num_workers=pool.max_count,
        )
        display(result)
        return True
    except Exception as e:
        display(e)
        print(f"failed to start workerpool {pool.name}, please start the pool manually")
        return False


def upgrade_prebuilt_image(
    client: SyftClient,
    old_image: SyftWorkerImage,
    mode: str = "manual",
) -> SyftWorkerImage | None:
    print(f"Found outdated prebuilt worker image `{old_image.image_identifier}`")
    if mode == "auto":
        new_syft_version = client.metadata.syft_version  # type: ignore
        new_identifier = upgrade_image_identifier(
            old_image.image_identifier, new_syft_version
        )
        new_image_tag = new_identifier.full_name_with_tag
    else:
        new_image_tag_or_none = get_tag_from_input()
        if not new_image_tag_or_none:
            return None
        new_image_tag = new_image_tag_or_none

    new_config = sy.PrebuiltWorkerConfig(
        tag=new_image_tag, description=old_image.config.description
    )

    print("submitting new prebuilt image...")
    try:
        result = client.api.services.worker_image.submit(worker_config=new_config)
        display(result)
        return result.value
    except Exception as e:
        print("could not submit new image")
        display(e)
        return None


def upgrade_syft_image(
    client: SyftClient,
    old_image: SyftWorkerImage,
    mode: str = "manual",
) -> SyftWorkerImage | None:
    old_identifier = old_image.image_identifier
    old_config = old_image.config
    new_syft_version = client.metadata.syft_version  # type: ignore

    if old_identifier is None:
        raise ValueError("old image does not have an image identifier")

    print(f"Found outdated custom worker image `{old_image.image_identifier}`")

    new_dockerfile = update_dockerfile_baseimage_tag(
        old_config.dockerfile, new_syft_version
    )

    if mode == "manual":
        confirm = confirm_dockerfile_update(old_config.dockerfile, new_dockerfile)
        if not confirm:
            return None

    # NOTE do not copy filename, it does not match the new dockerfile
    new_config = sy.DockerWorkerConfig(
        dockerfile=new_dockerfile, description=old_config.description, file_name=None
    )
    new_identifier = upgrade_image_identifier(old_identifier, new_syft_version)
    print(
        f"Updating image tag from {old_identifier.repo_with_tag} to {new_identifier.repo_with_tag}"
    )

    print("submitting new image...")
    try:
        submit_result = client.api.services.worker_image.submit(
            worker_config=new_config
        )
        custom_image = submit_result.value
    except Exception as e:
        print("could not submit new image")
        display(e)
        return None

    print("building new image...")
    try:
        client.api.services.worker_image.build(
            image_uid=custom_image.id,
            tag=new_identifier.repo_with_tag,
        )
    except Exception as e:
        print("could not build new image")
        display(e)
        return None

    return custom_image


def get_tag_from_input() -> str | None:
    new_image_tag = input(
        "Please enter the tag for the upgraded image. Type 'skip' to skip upgrading this workerpool"
    )
    if new_image_tag.lower() == "skip":
        return None
    return new_image_tag


def update_dockerfile_baseimage_tag(old_dockerfile: str, new_tag: str) -> str:
    is_updated = False
    new_dockerfile_ = []
    for line in old_dockerfile.splitlines():
        if line.startswith("FROM openmined/syft-backend:"):
            updated_line = f"FROM openmined/syft-backend:{new_tag}"
            new_dockerfile_.append(updated_line)
            is_updated = True
        else:
            new_dockerfile_.append(line)

    if not is_updated:
        raise ValueError("Could not update baseimage")
    return "\n".join(new_dockerfile_)


def confirm_dockerfile_update(old_dockerfile: str, new_dockerfile: str) -> bool:
    print("updated your dockerfile baseimage:")
    print("- Old dockerfile ----")
    print(old_dockerfile)
    print("- New dockerfile ----")
    print(new_dockerfile)
    print("---------------------")
    confirmation = input("is this correct? [y/n]")
    if confirmation.lower() not in ["y", "n"]:
        return confirm_dockerfile_update(old_dockerfile, new_dockerfile)
    return confirmation.lower() == "y"


def upgrade_image_identifier(
    old_identifier: SyftWorkerImageIdentifier, new_tag: str
) -> SyftWorkerImageIdentifier:
    return SyftWorkerImageIdentifier(
        registry=old_identifier.registry, repo=old_identifier.repo, tag=new_tag
    )



# stdlib

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...types.errors import SyftException
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from .image_registry import SyftImageRegistry
from .image_registry_stash import SyftImageRegistryStash

__all__ = ["SyftImageRegistryService"]


@serializable(canonical_name="SyftImageRegistryService", version=1)
class SyftImageRegistryService(AbstractService):
    stash: SyftImageRegistryStash

    def __init__(self, store: DBManager) -> None:
        self.stash = SyftImageRegistryStash(store=store)

    @service_method(
        path="image_registry.add",
        name="add",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def add(
        self,
        context: AuthedServiceContext,
        url: str,
    ) -> SyftSuccess:
        try:
            registry = SyftImageRegistry.from_url(url)
        except Exception as e:
            raise SyftException(public_message=f"Failed to create registry. {e}")

        stored_registry = self.stash.set(context.credentials, registry).unwrap()
        return SyftSuccess(
            message=f"Image Registry ID: {registry.id} created successfully",
            value=stored_registry,
        )

    @service_method(
        path="image_registry.delete",
        name="delete",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def delete(
        self,
        context: AuthedServiceContext,
        uid: UID | None = None,
        url: str | None = None,
    ) -> SyftSuccess:
        # TODO - we need to make sure that there are no workers running an image bound to this registry

        # if url is provided, get uid from url
        if url:
            self.stash.delete_by_url(context.credentials, url).unwrap()
            return SyftSuccess(
                message=f"Image Registry URL: {url} successfully deleted."
            )

        # if uid is provided, delete by uid
        if uid:
            self.stash.delete_by_uid(context.credentials, uid).unwrap()
            return SyftSuccess(
                message=f"Image Registry ID: {uid} successfully deleted."
            )
        else:
            raise SyftException(message="Either UID or URL must be provided.")

    @service_method(
        path="image_registry.get_all",
        name="get_all",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_all(
        self,
        context: AuthedServiceContext,
    ) -> list[SyftImageRegistry]:
        return self.stash.get_all(context.credentials).unwrap()

    @service_method(
        path="image_registry.get_by_id",
        name="get_by_id",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_by_id(self, context: AuthedServiceContext, uid: UID) -> SyftImageRegistry:
        return self.stash.get_by_uid(context.credentials, uid).unwrap()


TYPE_TO_SERVICE[SyftImageRegistry] = SyftImageRegistryService
SERVICE_TO_TYPES[SyftImageRegistryService].update({SyftImageRegistry})



# future
from __future__ import annotations

# stdlib
from collections.abc import Callable

# third party
from typing_extensions import Self

# relative
from ..abstract_server import AbstractServer
from ..abstract_server import ServerSideType
from ..abstract_server import ServerType
from ..deployment_type import DeploymentType
from ..serde.serializable import serializable
from ..server.credentials import SyftSigningKey
from ..service.queue.base_queue import QueueConfig
from ..store.blob_storage import BlobStorageConfig
from ..store.db.db import DBConfig
from ..store.document_store import StoreConfig
from ..types.syft_migration import migrate
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SYFT_OBJECT_VERSION_2
from ..types.syft_object import SyftObject
from ..types.transforms import TransformContext
from ..types.transforms import drop
from ..types.uid import UID


@serializable()
class WorkerSettings(SyftObject):
    __canonical_name__ = "WorkerSettings"
    __version__ = SYFT_OBJECT_VERSION_2

    id: UID
    name: str
    server_type: ServerType
    server_side_type: ServerSideType
    deployment_type: DeploymentType = DeploymentType.REMOTE
    signing_key: SyftSigningKey
    db_config: DBConfig
    blob_store_config: BlobStorageConfig | None = None
    queue_config: QueueConfig | None = None
    log_level: int | None = None

    @classmethod
    def from_server(cls, server: AbstractServer) -> Self:
        server_side_type = server.server_side_type or ServerSideType.HIGH_SIDE
        return cls(
            id=server.id,
            name=server.name,
            server_type=server.server_type,
            signing_key=server.signing_key,
            db_config=server.db_config,
            server_side_type=server_side_type,
            blob_store_config=server.blob_store_config,
            queue_config=server.queue_config,
            log_level=server.log_level,
            deployment_type=server.deployment_type,
        )


@serializable()
class WorkerSettingsV1(SyftObject):
    __canonical_name__ = "WorkerSettings"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    name: str
    server_type: ServerType
    server_side_type: ServerSideType
    deployment_type: DeploymentType = DeploymentType.REMOTE
    signing_key: SyftSigningKey
    document_store_config: StoreConfig
    action_store_config: StoreConfig
    blob_store_config: BlobStorageConfig | None = None
    queue_config: QueueConfig | None = None
    log_level: int | None = None


def set_db_config(context: TransformContext) -> TransformContext:
    if context.output:
        context.output["db_config"] = (
            context.server.db_config if context.server is not None else DBConfig()
        )
    return context


@migrate(WorkerSettingsV1, WorkerSettings)
def migrate_workersettings_v1_to_v2() -> list[Callable]:
    return [
        drop("document_store_config"),
        drop("action_store_config"),
        set_db_config,
    ]



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...store.linked_obj import LinkedObject
from ...types.result import as_result
from ...types.uid import UID
from .notifications import Notification
from .notifications import NotificationStatus


@serializable(canonical_name="NotificationSQLStash", version=1)
class NotificationStash(ObjectStash[Notification]):
    @as_result(StashException)
    def get_all_inbox_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> list[Notification]:
        if not isinstance(verify_key, SyftVerifyKey | str):
            raise AttributeError("verify_key must be of type SyftVerifyKey or str")
        return self.get_all(
            credentials,
            filters={"to_user_verify_key": verify_key},
        ).unwrap()

    @as_result(StashException)
    def get_all_sent_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> list[Notification]:
        if not isinstance(verify_key, SyftVerifyKey | str):
            raise AttributeError("verify_key must be of type SyftVerifyKey or str")
        return self.get_all(
            credentials,
            filters={"from_user_verify_key": verify_key},
        ).unwrap()

    @as_result(StashException)
    def get_all_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> list[Notification]:
        if not isinstance(verify_key, SyftVerifyKey | str):
            raise AttributeError("verify_key must be of type SyftVerifyKey or str")
        return self.get_all(
            credentials,
            filters={"from_user_verify_key": verify_key},
        ).unwrap()

    @as_result(StashException)
    def get_all_by_verify_key_for_status(
        self,
        credentials: SyftVerifyKey,
        verify_key: SyftVerifyKey,
        status: NotificationStatus,
    ) -> list[Notification]:
        if not isinstance(verify_key, SyftVerifyKey | str):
            raise AttributeError("verify_key must be of type SyftVerifyKey or str")
        return self.get_all(
            credentials,
            filters={
                "to_user_verify_key": str(verify_key),
                "status": status.name,
            },
        ).unwrap()

    @as_result(StashException, NotFoundException)
    def get_notification_for_linked_obj(
        self,
        credentials: SyftVerifyKey,
        linked_obj: LinkedObject,
    ) -> Notification:
        return self.get_one(
            credentials,
            filters={
                "linked_obj.id": linked_obj.id,
            },
        ).unwrap()

    @as_result(StashException, NotFoundException)
    def update_notification_status(
        self, credentials: SyftVerifyKey, uid: UID, status: NotificationStatus
    ) -> Notification:
        notification = self.get_by_uid(credentials, uid=uid).unwrap()
        notification.status = status
        return self.update(credentials, obj=notification).unwrap()

    @as_result(StashException, NotFoundException)
    def delete_all_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> bool:
        if not isinstance(verify_key, SyftVerifyKey | str):
            raise AttributeError("verify_key must be of type SyftVerifyKey or str")
        notifications = self.get_all_inbox_for_verify_key(
            credentials,
            verify_key=verify_key,
        ).unwrap()
        for notification in notifications:
            self.delete_by_uid(credentials, uid=notification.id).unwrap()
        return True
